/*
Tangle & Whisper
Tag Team Swap Mechanic from Silverhorn
Used with Permission
	by SMS Alfredo
	& Lach
*/

local SKIN1 = "tangle"
local SKIN2 = "whisper"

if not sfx_pswap then freeslot("sfx_pswap") end
if not sfx_pfail then freeslot("sfx_pfail") end

sfxinfo[sfx_pswap].caption = "Switch!"
sfxinfo[sfx_pfail].caption = "Can't Switch!"

// if a player is in any of these panims, they will be set to S_PLAY_STND instead
local PANIM_BLACKLIST = {
	PA_ETC,
	PA_DASH,
	PA_ABILITY,
	PA_ABILITY2
}

// swap these pflags, leave the rest alone
local PFLAG_WHITELIST = PF_JUMPED|PF_NOJUMPDAMAGE|PF_THOKKED|PF_SPINNING|PF_STARTDASH|PF_SHIELDABILITY|PF_GLIDING|PF_BOUNCING|PF_CANCARRY

// @s1 and @s2 can be players, mobjs, tables, etc.
// @field is a string, the name of the field to be swapped
local function SwapStructField(s1, s2, field)
	s1[field], s2[field] = $2, $1
end

addHook("MapLoad", do
	local player1 = players[0]
	local mo1 = player1 and player1.mo
	local player2 = players[1]
	local mo2 = player2 and player2.mo
	
	if not netgame and not multiplayer
	and mo1 and (mo1.skin == "tangle" or mo1.skin == "whisper")
	and mo2 and (mo2.skin == "tangle" or mo2.skin == "whisper")
	and mo1.skin == mo2.skin
		if mo1.skin == "whisper"
			R_SetPlayerSkin(player2, "tangle")
		else
			R_SetPlayerSkin(player2, "whisper")
		end
	end
end)

addHook("PreThinkFrame", do
	if netgame or multiplayer
		return
	end
	
	local player1 = players[0]
	local mo1 = player1 and player1.mo
	local player2 = players[1]
	local mo2 = player2 and player2.mo
	
	if mo1 and (mo1.skin == "tangle" or mo1.skin == "whisper")
		if not (player1.cmd.buttons & BT_CUSTOM1)
		or player1.pflags & PF_STASIS or player1.exiting
			player1.whisperswapready = true
			player1.whisperswapping = false
		elseif player1.whisperswapready then
			player1.whisperswapping = true
			player1.whisperswapready = false
		else
			player1.whisperswapping = false
		end
		
		if not (mo2 and (mo2.skin == "tangle" or mo2.skin == "whisper"))
		then return end
		
		player1.whispercustom = nil
		player2.whispercustom = nil
		
		if P_PlayerInPain(player1) or player1.powers[pw_carry] or not mo1.health
		or (mo1.state >= S_PLAY_SUPER_TRANS1 and mo1.state <= S_PLAY_SUPER_TRANS6)
		or (player1.solchar and player1.solchar.istransformed == 1)
		or P_PlayerInPain(player2) or player2.powers[pw_carry] or not mo2.health
		or (mo2.state >= S_PLAY_SUPER_TRANS1 and mo2.state <= S_PLAY_SUPER_TRANS6)
		or (player2.solchar and player2.solchar.istransformed == 1)
			if player1.whisperswapping
			and not mapheaderinfo[gamemap].silverhorn_endingmap
				S_StartSound(mo1, sfx_pfail, player1)
			end
			return
		end
		
		// set super
		if player1.powers[pw_super]
			if mo1.state == S_PLAY_SUPER_TRANS1
			and mo2.state ~= S_PLAY_SUPER_TRANS1
				mo2.state = S_PLAY_SUPER_TRANS1
			end
			player2.charflags = $ | SF_SUPER
			player2.powers[pw_super] = player1.powers[pw_super]
			player2.rings = player1.rings
		end
		
		// set sol
		if player1.solchar and player1.solchar.istransformed and player2.solchar
			player2.solemeralds = player1.solemeralds
			player2.rings = player1.rings
			if not player2.solchar.istransformed
				Silverhorn.SolTransformation(player2, player1.solchar.istransformed == 1)
			end
		end
		
		// match p2's commands to p1
		player2.tangleswap = player1.tangleswap
		player2.whisperswap = player1.whisperswap
		player2.whisperwisptoggle = player1.whisperwisptoggle
		player2.whispercamera = player1.whispercamera
		player2.whisperaim = player1.whisperaim
		
		// bot command stuff
		if player2.bot == 1
			// mimic player 1 during sol transformation
			if player2.solchar and player2.solchar.istransformed == 1
				player2.cmd.buttons = player1.cmd.buttons
			end
			
			// no jumping allowed while being carried!
			if player2.powers[pw_carry] == CR_PLAYER
				player2.cmd.buttons = $ &~ BT_JUMP
			end
		end
		
		// start a swap
		if player1.whisperswapping and not transforming
			S_StartSound(mo1, sfx_pswap)
			
			// swap colors
			SwapStructField(player1, player2, "skincolor")
			SwapStructField(mo1, mo2, "color")
			
			// swap skins
			local xmomSpeed
			if player1.isxmomentum
				xmomSpeed = player1.normalspeed
			end
			
			SwapStructField(mo1, mo2, "skin")
			R_SetPlayerSkin(player1, mo1.skin)
			R_SetPlayerSkin(player2, mo2.skin)
			
			if xmomSpeed ~= nil
				player1.normalspeed = xmomSpeed
				player2.normalspeed = xmomSpeed
			end
			
			// whisper-specific variables
			SwapStructField(mo1, mo2, "whispermask")
			SwapStructField(player1, player2, "whispermask")
			SwapStructField(player1, player2, "whisperguncolor")
			SwapStructField(player1, player2, "whisperwisp")
			for i = 0, 1
				local player = players[i]
				if player.whisperwisp and player.whisperwisp.valid
					player.whisperwisp.target = player.mo
				end
			end
			
			// cancel BattleMod abilities
			player1.actionstate = 0
			player1.actiontime = 0
			player2.actionstate = 0
			player2.actiontime = 0
			
			// swap state
			SwapStructField(mo1, mo2, "state")
			
			// swap pflags
			local pflags = player1.pflags & PFLAG_WHITELIST
			player1.pflags = $ & ~PFLAG_WHITELIST | (player2.pflags & PFLAG_WHITELIST)
			player2.pflags = $ & ~PFLAG_WHITELIST | pflags
			
			// don't retain PF_JUMPED on the ground!
			if player1.pflags & PF_JUMPED and P_IsObjectOnGround(mo1)
				player1.pflags = $ & ~PF_JUMPED
			end
			if player2.pflags & PF_JUMPED and P_IsObjectOnGround(mo2)
				player2.pflags = $ & ~PF_JUMPED
			end

			P_MoveOrigin(mo2, mo1.x, mo1.y, mo1.z)
			mo2.momx = mo1.momx
			mo2.momy = mo1.momy
			P_Thrust(mo2, player1.drawangle,
			min(-8*mo1.scale,FixedMul(player1.speed/-2,mo1.scale)))
			mo2.momz = mo1.momz
			mo2.angle = mo1.angle
			player2.drawangle = player1.drawangle
			
			// prevent raising errors when switching into certain states
			for i = 0, 1
				local player = players[i]
				local mo = player.mo
				for _, panim in ipairs(PANIM_BLACKLIST)
					if player.panim == panim
						if player.pflags & PF_SPINNING
						or (player.pflags & PF_JUMPED and not (player.pflags & PF_NOJUMPDAMAGE))
							mo.state = S_PLAY_ROLL
						elseif P_IsObjectOnGround(mo)
							mo.state = S_PLAY_STND
						else
							mo.state = S_PLAY_FALL
						end
					end
				end
			end
		end
	end
end)